// ========================================================================
// Copyright 2006-2007 Mort Bay Consulting Pty. Ltd.
// ------------------------------------------------------------------------
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ========================================================================

package org.mortbay.jetty.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.util.concurrent.CountDownLatch;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import junit.framework.TestCase;

import org.mortbay.io.Buffer;
import org.mortbay.io.ByteArrayBuffer;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.EofException;
import org.mortbay.jetty.HttpConnection;
import org.mortbay.jetty.HttpMethods;
import org.mortbay.jetty.Request;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.AbstractHandler;
import org.mortbay.jetty.nio.SelectChannelConnector;

/**
 * Functional testing for HttpExchange.
 * 
 * @author Matthew Purland
 */
public class HttpExchangeTest extends TestCase
{
    private Server _server;
    private int _port;
    private HttpClient _connectionPool;

    protected void setUp() throws Exception
    {
        startServer();
        _connectionPool=new HttpClient();
        _connectionPool.setConnectorType(HttpClient.CONNECTOR_SELECT_CHANNEL);
        _connectionPool.start();
    }

    protected void tearDown() throws Exception
    {
        stopServer();
    }

    public void testPerf() throws Exception
    {
        testSend(1);
        
        testSend(10);

        /*
        for (int i=100; i<=10000; i*=10)
        {
            testSend(i);
        }
        */
    }

    /**
     * Test sending data through the exchange.
     * 
     * @throws IOException
     */
    public void testSend(final int nb) throws Exception
    {
        final CountDownLatch latch=new CountDownLatch(nb);
        long l0=System.currentTimeMillis();
        for (int i=0; i<nb; i++)
        {
            HttpExchange httpExchange=new HttpExchange()
            {
                protected void onRequestCommitted()
                {
                    System.err.println("Request committed");
                }

                protected void onResponseStatus(Buffer version, int status, Buffer reason)
                {
                    System.err.println("Response Status: " + version+" "+status+" "+reason);
                }

                protected void onResponsetHeader(Buffer name, Buffer value)
                {
                    System.err.println("Response header: " + name + " = " + value);
                }

                protected void onResponseContent(Buffer content)
                {
                    System.err.println("Response content:" + content);
                }

                protected void onResponseComplete()
                {
                    System.err.println("Response completed");
                    latch.countDown();
                }
            };

            httpExchange.setURL("http://localhost:"+_port+"/");
            httpExchange.addRequestHeader("arbitrary","value");
            
            _connectionPool.send(httpExchange);
        }
        
        // System.err.println("Messages sent");
        latch.await();
        long l1=System.currentTimeMillis();
        System.err.println("Delay (jetty, "+nb+"): "+(l1-l0)+" ms");
    }

    public void testPostWithStreamedResponse() throws Exception
    {
        HttpExchange.StreamedExchange httpExchange=new HttpExchange.StreamedExchange();
        httpExchange.setURL("http://localhost:"+_port+"/");
        httpExchange.setMethod(HttpMethods.POST);
        httpExchange.setRequestContent(new ByteArrayBuffer("<hello />"));
        _connectionPool.send(httpExchange);
        httpExchange.waitForStatus(HttpExchange.STATUS_PARSING_CONTENT);
        InputStream is=httpExchange.getResponseStream();
        copyStream(is,System.err);
    }

    public void testGetWithStreamedResponse() throws Exception
    {
        HttpExchange.StreamedExchange httpExchange=new HttpExchange.StreamedExchange();
        httpExchange.setURL("http://localhost:"+_port+"/");
        httpExchange.setMethod(HttpMethods.GET);
        _connectionPool.send(httpExchange);
        httpExchange.waitForStatus(HttpExchange.STATUS_PARSING_CONTENT);
        InputStream is=httpExchange.getResponseStream();
        copyStream(is,System.err);
        is.close();
    }

    public static void copyStream(InputStream in, OutputStream out)
    {
        try
        {
            byte[] buffer=new byte[1024];
            int len;
            while ((len=in.read(buffer))>=0)
            {
                out.write(buffer,0,len);
            }
        }
        catch (EofException e)
        {
            
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    private void startServer() throws Exception
    {
        _server=new Server();
        Connector connector=new SelectChannelConnector();
        connector.setPort(0);
        _server.setConnectors(new Connector[]
        { connector });
        _server.setHandler(new AbstractHandler()
        {
            public void handle(String target, HttpServletRequest request, HttpServletResponse response, int dispatch) throws IOException, ServletException
            {
                Request base_request=(request instanceof Request)?(Request)request:HttpConnection.getCurrentConnection().getRequest();
                base_request.setHandled(true);
                response.setStatus(200);
                if (request.getMethod().equalsIgnoreCase("GET"))
                {
                    response.getOutputStream().println("<hello>");
                    for (int i=0; i<100; i++)
                    {
                        response.getOutputStream().println("  <world>"+i+"</world");
                        if (i%20==0)
                            response.getOutputStream().flush();
                    }
                    response.getOutputStream().println("</hello>");
                }
                else
                {
                    copyStream(request.getInputStream(),response.getOutputStream());
                }
            }
        });
        _server.start();
        _port=connector.getLocalPort();
    }

    private void stopServer() throws Exception
    {
        _server.stop();
        Thread.sleep(1000);
    }
}
